@@ -6,21 +6,22 @@
 ## This code is free software; you can redistribute it and/or modify
 ## it under the same terms as Perl itself.
 ##
-## $Id: Agent.pm,v 1.73 2003/05/07 00:00:38 vipul Exp $
+## $Id: Agent.pm,v 1.96 2006/05/27 00:00:53 rsoderberg Exp $
 
 package Razor2::Client::Agent;
 
 use lib qw(lib);
 use strict;
-use Getopt::Long; 
+use Getopt::Long;
 use IO::File;
 
+use Razor2::String qw(fisher_yates_shuffle);
+
 use base qw(Razor2::Client::Core);
 use base qw(Razor2::Client::Config);
 use base qw(Razor2::Logger);
-use base qw(Razor2::String); 
+use base qw(Razor2::String);
 use Razor2::Preproc::Manager;
-use Razor2::Preproc::VR1;
 use Data::Dumper;
 use vars qw( $VERSION $PROTOCOL );
 
@@ -38,9 +39,9 @@ sub new {
     delete $ENV{BASH_ENV};
 
     my @valid_program_names = qw(
-            razor-check 
-            razor-report  
-            razor-revoke  
+            razor-check
+            razor-report
+            razor-revoke
             razor-admin
     );
 
@@ -48,22 +49,23 @@ sub new {
     foreach (@valid_program_names) { $breed =~ /$_$/ and $ok = $_; }
     unless ($ok) {
         if ($breed =~ /razor-client$/) {
-            #create_symlinks(@valid_program_names);
+            # We no longer create symlinks, but for backwards compatibility
+            # return success.
             exit 0;
         }
         die "Invalid program name, must be one of: @valid_program_names\n";
     }
 
-    $ok =~ /razor-(.*)$/; 
+    $ok =~ /razor-(.*)$/;
     my %me = (
             name_version => "Razor-Agents v$VERSION",  # used in register
             breed        => $1,
-            preproc_vr1  => new Razor2::Preproc::VR1,
-            preproc      => new Razor2::Preproc::Manager,
+            preproc      => new Razor2::Preproc::Manager (no_deHTMLcomment => 1),
+            preproc_vr8  => new Razor2::Preproc::Manager (no_deHTML => 1),
             global_razorhome => '/etc/razor',
     );
-    
-    
+
+
     return bless \%me, $class;
 }
 
@@ -72,7 +74,7 @@ sub do_conf {
 
     # parse config-related cmd-line args
     #
-    
+
     # identity is parsed later after razorhome is fully resolved
 
     if ($self->{opt}->{config}) {
@@ -81,16 +83,18 @@ sub do_conf {
         } elsif (-r $self->{opt}->{config}) {
             $self->{razorconf} = $self->{opt}->{config};
         } else {
-            return $self->error("Can't read conf file: $self->{opt}->{config}") 
+            return $self->error("Can't read conf file: $self->{opt}->{config}")
         }
     }
     if ($self->{opt}->{razorhome}) {
         if (-d $self->{opt}->{razorhome}) {
             $self->{razorhome} = $self->{opt}->{razorhome};
         } else {
-            return $self->error("Can't read: $self->{opt}->{razorhome}") 
+            return $self->error("Can't read: $self->{opt}->{razorhome}")
                 unless $self->{opt}->{create_conf};
         }
+        # once razorhome is successfully overridden, override the global razorhome as well.
+        $self->{global_razorhome} = $self->{razorhome};
     }
     return unless $self->read_conf();
 
@@ -105,6 +109,7 @@ sub do_conf {
         $self->log(8," -discover will force complete discovery");
     }
     if ($self->{opt}->{debug} && !$self->{opt}->{debuglevel}) {
+        $self->{conf}->{debuglevel} ||= 9;
         $self->{conf}->{debuglevel} = 9 if $self->{conf}->{debuglevel} < 9;
     }
 
@@ -113,10 +118,20 @@ sub do_conf {
     # Note: we start logging before we process '-create' ,
     # so logfile will not go into a newly created razorhome
     #
-    my $logto = $self->{opt}->{debug} ? "stdout" : "file:$self->{conf}->{logfile}";
+    #my $logto = $self->{opt}->{debug} ? "stdout" : "file:$self->{conf}->{logfile}";
+    my $logto;
+    if ($self->{opt}->{debug}) {
+        $logto = 'stdout';
+    } elsif ($self->{conf}->{logfile} eq 'syslog') {
+        $logto = 'syslog';
+    } elsif ($self->{conf}->{logfile} eq 'sys-syslog') {
+        $logto = 'sys-syslog';
+    } else {
+        $logto = "file:$self->{conf}->{logfile}";
+    }
     if (exists $self->{conf}->{logfile}) {
         my $debuglevel = exists $self->{conf}->{debuglevel} ? $self->{conf}->{debuglevel} : 9;
-        my $logger = new Razor2::Logger ( 
+        my $logger = new Razor2::Logger (
                         LogDebugLevel => $debuglevel,
                         LogTo         => $logto,
                         LogPrefix     => $self->{breed},
@@ -127,16 +142,13 @@ sub do_conf {
         $self->{logref} = ref($logger) ? $logger : 0;
         # log error strings at loglevel 11.  Pick a high number 'cuz
         # if its really an error, it will be in errstr for caller
-        $self->{logerrors} = 11; 
+        $self->{logerrors} = 11;
     }
     $self->logobj(15,"cmd-line options", $self->{opt});
     $self->{preproc}->{rm}->{log} = $self->{logref};
 
-    #my $chadrules;  # test warning
-    #$self->logobj(3,"warn on this $chadrules");
-
     # creates razorhome, and sets $self->{razorhome} if successful
-    return $self->errprefix("Could not create 'razorhome'") unless $self->create_home_conf(); 
+    return $self->errprefix("Could not create 'razorhome'") unless $self->create_home_conf();
     $self->compute_identity;
 
     $self->log(5,"computed razorhome=$self->{razorhome}, conf=$self->{razorconf}, ident=$self->{identity}");
@@ -156,7 +168,7 @@ sub create_home_conf {
 
     unless ($self->{opt}->{create_conf}) {
         #
-        # if the global razorhome exists, don't create anything 
+        # if the global razorhome exists, don't create anything
         # without '-create' option
         #
         return 1 if (-d $self->{global_razorhome});
@@ -174,8 +186,8 @@ sub create_home_conf {
     #
     # user passed in 'create' option, so create.
     #
-    my $rhome = $self->{opt}->{razorhome} 
-              ? $self->{opt}->{razorhome} 
+    my $rhome = $self->{opt}->{razorhome}
+              ? $self->{opt}->{razorhome}
               : $self->{razorhome_computed};
 
     if ($rhome) {
@@ -193,7 +205,7 @@ sub create_home_conf {
         # if create and conf specified, exit if write is not successful
         #
         $self->{razorconf} = $self->{opt}->{config};
-        return $self->write_conf(); 
+        return $self->write_conf();
 
     } else {
 
@@ -216,7 +228,6 @@ sub log {
     if ($self->{logref}) {
         return  $self->{logref}->log($level, $msg);
     } elsif ($self->{opt}->{debug}) {
-    #} else {
         print " Razor-Log: $msg\n" if $self->{opt}->{debug};
     }
 }
@@ -277,7 +288,7 @@ EOFSIGS
    -M | --mbox  Accept a mailbox name on the command line (default)
                 If no filename, mbox, or signatures, input read from stdin.
 EOFMBOX
-    
+
     my %b;
     $b{check} = <<EOFCHECK;
 
@@ -344,7 +355,7 @@ EOFFUTURE
 
 # maybe this should be in Client::Config
 #
-sub read_options { 
+sub read_options {
     my ($self, $agent) = @_;
     $self->{args} = join ' ', @ARGV;
     Getopt::Long::Configure ("no_ignore_case");
@@ -353,7 +364,7 @@ sub read_options {
     # These options override what is loaded in config file
     # the names on the right should match keys in config file
     #
-    GetOptions(
+    my $ret = GetOptions(
         's'   => \$opt{simulate},
         'd'   => \$opt{debug},
   'verbose'   => \$opt{debug},
@@ -395,28 +406,33 @@ sub read_options {
      'le=s'   => \$opt{logic_engines},
     );
 
+    if ($ret == 0) {
+        $self->error("failed to parse command line options.\n");
+        return;
+    }
+
     # remove elements not set in the cmd-line
     foreach (keys %opt) { delete $opt{$_} unless defined $opt{$_}; }
 
-    if ($opt{usage}) { 
+    if ($opt{usage}) {
         $self->error($self->_help);
         return;
-    } elsif ($opt{mbox} && $opt{sig}) { 
-        $self->error("--mbox and --sig are mutually exclusive.\n"); 
+    } elsif ($opt{mbox} && $opt{sig}) {
+        $self->error("--mbox and --sig are mutually exclusive.\n");
         return;
-    } elsif ($opt{sig} && !$opt{sigengine}) { 
+    } elsif ($opt{sig} && !$opt{sigengine}) {
         $self->error("--sig requires -e (engine used to generate sig)\n");
         return;
         #
         # fixme - require ep4 if -e 4 is used ?
         #
-    } elsif ($opt{version}) { 
+    } elsif ($opt{version}) {
         $self->error("Razor Agents $VERSION, protocol version $PROTOCOL");
         return;
-    } 
+    }
     $self->{opt} = \%opt;
     return 1;
-} 
+}
 
 
 
@@ -429,23 +445,23 @@ sub checkit {
     my $args = shift;
 
     # check for spam.
-    # input can be one of 
+    # input can be one of
     #   file - single mail
     #   mbox - many  mail
     #   sig  - 1 or more sigs
     #   or a filehandle provided via args
-        
+
     my $objects;
     if ($self->{conf}->{sig}) {
         my @sigs;
-        # 
+        #
         # cmd-line sigs
-        # 
-        # prepare 1 mail object per sig 
-        # 
+        #
+        # prepare 1 mail object per sig
+        #
         foreach my $sig (split ',', $self->{conf}->{sig}) {
             $sig =~ s/^\s*//;  $sig =~ s/\s*$//;
-            my $hr = { 
+            my $hr = {
                 eng => $self->{conf}->{sigengine},
                 sig => $sig,
             };
@@ -456,10 +472,11 @@ sub checkit {
         $self->log (5,"received ". (scalar @sigs) ." valid cmd-line sigs.");
         $objects = $self->prepare_objects(\@sigs) or return 2;
     } else {
-        
+
         my $mails = $self->parse_mbox($args) or return 2;
 
         $objects  = $self->prepare_objects($mails) or return 2;
+
         #
         # if mail is whitelisted, its not spam.
         # flag it so it we don't check it against server
@@ -479,7 +496,7 @@ sub checkit {
     $self->get_server_info()                            or return 2;
     my $printable_sigs = $self->compute_sigs($objects)  or return 2;
 
-    if ($self->{opt}->{printhash}) { 
+    if ($self->{opt}->{printhash}) {
         my $i = 0;
         foreach (@$printable_sigs) {
             if ($self->{opt}->{sigengine}) {
@@ -490,7 +507,7 @@ sub checkit {
         }
         $self->log (4, "Done. Printed $i sig(s) for ". scalar(@$objects) ." mail(s)");
     }
-    if ($self->{opt}->{printcleaned}) { 
+    if ($self->{opt}->{printcleaned}) {
         my $totalp = 0;
         my $totalc = 0;
         foreach my $obj (@$objects) {
@@ -512,13 +529,13 @@ sub checkit {
         $self->log (4, "Done. $totalp uncleaned, $totalc cleaned mails saved in $self->{opt}->{printcleaned}");
         print "Done. $totalp uncleaned, $totalc cleaned mails saved in $self->{opt}->{printcleaned}\n";
         return 1;
-            
+
     }
 
     return 1 if $self->{opt}->{printhash};
 
     # only check good objects
-    my @goodones;                 # this should be optimized! 
+    my @goodones;                 # this should be optimized!
     foreach my $obj (@$objects) {
         next if $obj->{skipme};
         push @goodones, $obj;
@@ -532,17 +549,24 @@ sub checkit {
         $self->log (4, "Done. (simulate only)");
         return 1;
     }
+
+    #
+    # Connect to catalogue server
+    #
+    $self->{s}->{list} = $self->{s}->{catalogue};
+    $self->nextserver();
+    $self->connect()          or return 2;
+
     #
     # Check against server
-    #  
-    $self->connect()          or return 2;
+    #
     $self->check (\@goodones) or return 2;
     $self->disconnect()       or return 2;
 
 
     #
     # print out responses and exit
-    #  
+    #
     my $only1check = (scalar(@$objects) == 1) ? 1 : 0;
     my $has_spam = 0;
     foreach my $obj (@$objects) {
@@ -578,7 +602,7 @@ sub checkit {
 
 # returns 0 if success
 # returns 2 if error
-sub adminit { 
+sub adminit {
     my $self = shift;
 
     my $done_something = 0;
@@ -587,12 +611,12 @@ sub adminit {
         $done_something++;
         # $self->create_home_conf() is always checked
     }
-    
-    if (  $self->{opt}->{force_discovery} || 
+
+    if (  $self->{opt}->{force_discovery} ||
           $self->{opt}->{create_conf}) {
         $done_something++;
         # get_server_info() calls nextserver() which calls discovery()
-        $self->get_server_info()    or return 2;  
+        $self->get_server_info()    or return 2;
     }
 
     if ($self->{opt}->{register}) {
@@ -605,14 +629,14 @@ sub adminit {
         $self->error("An option needs to be specified,  -h for help.");
         return 2;
     }
-    
+
     return 0;
 }
 
 # returns 0 if success
 # returns 2 if error
-sub registerit { 
-    my $self = shift;
+sub registerit {
+    my($self, $auto) = @_;
 
     unless ($self->{razorhome} || $self->{opt}->{identity}) {
         $self->errprefix("Unable to register without a valid razorhome or identity");
@@ -621,31 +645,50 @@ sub registerit {
 
     my $ident;
 
-    if (exists $self->{opt}->{user} 
-        && ($ident = $self->get_ident) 
+    if (exists $self->{opt}->{user}
+        && ($ident = $self->get_ident)
         && $ident->{user} eq $self->{opt}->{user} ) {
         $self->error("You are already registered as user=$ident->{user} in $self->{razorhome}");
         return 2;
     }
-    if ( $self->{conf}->{simulate}) {
+    if ($self->{conf}->{simulate}) {
         $self->log(5,"Done - simulate only.");
         return 0;
     }
 
-    $self->get_server_info()    or return 2;
-    $self->connect()            or return 2;
     if ($self->{opt}->{create_conf}) {
-        $self->disconnect()     or return 2;
         $self->log(3, "Register create successful.");
         return 0;
     }
 
-    $ident = $self->register({
-                        user => $self->{opt}->{user}, 
-                        pass => $self->{opt}->{pass}, 
-                            }) or return 2;
-    $self->disconnect()         or return 2;
+    if ($auto) {
+        $self->log(3, "Write test underway");
+        my($ident) = {
+            user    =>  'writetest',
+            pass    =>  'writetest',
+        };
+        my($fn);
+        unless ($fn = $self->save_ident($ident)) {
+            $self->log(3, "Unable to write identity to home");
+            return 2;
+        }
+        unlink($fn) or return 2;
+        $self->log(3, "Write test completed");
+    }
 
+    $self->get_server_info()    or return 2;
+    $self->connect()            or return 2;
+
+    $self->log(3, "Attempting to register.");
+    # attempt to register the user/pass
+    $ident = $self->register_identity($self->{opt}->{user}, $self->{opt}->{pass});
+
+    $self->disconnect()     or return 2;
+
+    unless (ref $ident) {
+        $self->log(3, "Failed to register identity.");
+        return 2;
+    }
 
     if (my $fn = $self->save_ident($ident)) {
         my $msg = "Register successful.  Identity stored in $fn";
@@ -653,11 +696,11 @@ sub registerit {
         print "$msg\n";
         return 0;
     } else {
+        $self->log(3, "Register failed.");
         return 2;
     }
 }
 
-
 #
 # handles report and revoke
 #
@@ -669,24 +712,46 @@ sub reportit {
 
     my $ident = $self->get_ident;
     unless ($ident) {
-        $self->errprefix("Bootstrap Error: Your Razor2 identity was not found.\n   " . 
-                         "  If you haven't registered, please do so:\n" .
-                         "     \"razor-admin -register -user=[name/email_address] -pass=[password]\".\n". 
-                         "     (Further information can be found in the razor-admin(1) manpage)\n" .
-                         "  If you did register, please ensure your identity symlink (or file) is in order.\nmore");
-
-        return 2;
+        $self->log(3, "Razor2 identity not found.  Attempting to register automatically.");
+        if ($self->registerit("auto")) {
+            $self->log(3, "Automatic registration failed.");
+            $self->errprefix("Bootstrap Error: Your Razor2 identity was not found.\n   " .
+                             "  If you haven't registered, please do so:\n" .
+                             "     \"razor-admin -register -user=[name/email_address] -pass=[password]\".\n".
+                             "     (Further information can be found in the razor-admin(1) manpage)\n" .
+                             "  If you did register, please ensure your identity symlink (or file) is in order.\n");
+            return 2;
+        }
+        $ident = $self->get_ident;
+        unless ($ident) {
+            $self->log(3, "Unable to load automatically registered identity.");
+            $self->errprefix("Bootstrap Error: Your Razor2 identity was not found.\n   " .
+                             "  If you haven't registered, please do so:\n" .
+                             "     \"razor-admin -register -user=[name/email_address] -pass=[password]\".\n".
+                             "     (Further information can be found in the razor-admin(1) manpage)\n" .
+                             "  If you did register, please ensure your identity symlink (or file) is in order.\n");
+            return 2;
+        }
     }
 
-    # background myself 
+    if (!$self->{opt}{foreground} &&
+        (@ARGV < 1 || $ARGV[0] eq "-" || $ARGV[0] eq "")) {
+        if (-t STDIN) {
+            $self->error("Unable to read from a TTY using STDIN while forked. \n" .
+                         "Doing so leads to undefined behaviour in certain shells.");
+            return 2;
+        }
+    }
+
+    # background myself
     unless ($self->{opt}->{foreground}) {
-        chdir '/'; 
-        fork && return 0; 
+        chdir '/';
+        fork && return 0;
         POSIX::setsid;
         # close 0, 1, 2;
     }
 
-    if ($self->{opt}->{authen_only}) { 
+    if ($self->{opt}->{authen_only}) {
         $self->authenticate($ident) or return;
         $self->log(5,"Done - authenticate only.");
         return 0 if $self->{authenticated};
@@ -695,13 +760,15 @@ sub reportit {
 
     my $mails   = $self->parse_mbox($args) or return 2;
 
-    my $objects = $self->prepare_objects($mails) or return 2;  
+    my $objects = $self->prepare_objects($mails) or return 2;
+
 
     # compute_sigs needs server info like ep4, so get_server_info first
     $self->get_server_info()                            or return 2;
+
     my $printable_sigs = $self->compute_sigs($objects)  or return 2;
 
-    if ($self->{opt}->{printhash}) { 
+    if ($self->{opt}->{printhash}) {
         foreach (@$printable_sigs) {
             if ($self->{opt}->{sigengine}) {
                 next unless (/ e$self->{opt}->{sigengine}: /);
@@ -720,10 +787,12 @@ sub reportit {
         exit 1;
     }
 
+    $self->{s}->{list} = $self->{s}->{nomination};
+    $self->nextserver();
     $self->connect()            or return 2;
-    $self->authenticate($ident) or return 2; 
-    $self->report($objects)     or return 2; 
-    $self->disconnect()         or return 2; 
+    $self->authenticate($ident) or return 2;
+    $self->report($objects)     or return 2;
+    $self->disconnect()         or return 2;
 
 
     if ($self->{opt}->{foreground}) {
@@ -745,24 +814,24 @@ sub parse_mbox {
     my @message;
     my $passed_fh = 0;
     my $aref;
-    
+
     # There are different kinds of mbox formats, we just split on simplest case.
-    # djb defines mbox, mboxrd, mboxcl, mboxcl2 
+    # djb defines mbox, mboxrd, mboxcl, mboxcl2
     # http://www.qmail.org/qmail-manual-html/man5/mbox.html
     #
     # non-mbox support added, thanx to Aaron Hopkins <aaron@die.net>
 
-    if (exists $$args{"fh"}) { 
+    if (exists $$args{"fh"}) {
         @ARGV = ();
         push @ARGV, $$args{'fh'};
         $passed_fh = 1;
-    } elsif (exists $$args{"aref"}) { 
+    } elsif (exists $$args{"aref"}) {
        $aref = $$args{"aref"};
-    } elsif (!scalar @ARGV) { 
+    } elsif (!scalar @ARGV) {
         push @ARGV, "-"
     }
 
-    if ($$args{'aref'}) { 
+    if ($$args{'aref'}) {
         my @foo = (\join'', @{$$args{'aref'}});
         return \@foo;
     }
@@ -770,9 +839,9 @@ sub parse_mbox {
     foreach my $file (@ARGV) {
         my $fh = new IO::File;
         my @message = ();
-        if (ref $file) { 
+        if (ref $file) {
             $fh = $file
-        } else { 
+        } else {
             open $fh, "<$file" or return $self->error("Can't open $file: $!");
         }
 
@@ -780,7 +849,7 @@ sub parse_mbox {
         next unless $line;
 
         if ($line =~ /^From /) {
-            $self->log(8,"reading  mbox formatted mail from ". 
+            $self->log(8,"reading  mbox formatted mail from ".
                 ($file eq '-' ? "<stdin>" : $file));
             while (1) {
                 push @message, $line;
@@ -792,14 +861,14 @@ sub parse_mbox {
                 }
             }
         } else {
-            $self->log(8,"reading straight RFC822 mail from ". 
+            $self->log(8,"reading straight RFC822 mail from ".
                 ($file eq '-' ? "<stdin>" : $file));
             push @mails, \join ('', map {s/^(>*From )/>$1/; $_} $line, <$fh>);
         }
         close $fh unless $passed_fh;
     }
 
-    my $cnt = scalar @mails; 
+    my $cnt = scalar @mails;
     $self->log (6, "read $cnt mail". ($cnt>1 ? 's' : '') );
 
     return \@mails;
@@ -807,20 +876,20 @@ sub parse_mbox {
 
 
 
-sub raise_error { 
+sub raise_error {
     my ($self, $errstr) = @_;;
     my $str;
-    if (ref $self) { 
+    if (ref $self) {
         $str = $self->errstr;
-    } 
+    }
     $str = $errstr if $errstr;
     my ($code) = $str =~ /Razor Error (\d+):/;
-    $code = 255 unless $code; 
+    $code = 255 unless $code;
     print "FATAL: $str";
     exit $code;
 }
 
-# returns 1 if mail should be skipped 
+# returns 1 if mail should be skipped
 #
 sub local_check {
     my ($self, $obj) = @_;
@@ -828,8 +897,8 @@ sub local_check {
 
     $headers =~ s/\n\s+//sg;  # merge multi-line headers
 
-    if ($self->{conf}->{ignorelist}) { 
-        if ($headers =~ /\n((X-)?List-Id[^\n]+)/i) { 
+    if ($self->{conf}->{ignorelist}) {
+        if ($headers =~ /\n((X-)?List-Id[^\n]+)/i) {
             my $listid = $1;
             my ($line1) = substr(${$obj->{orig_mail}}, 0, 50) =~ /^([^\n]+)/;
             $self->log (5,"Mailing List post; mail ". $obj->{id} ." not spam.");
@@ -838,16 +907,16 @@ sub local_check {
         }
     }
     return 0 if $self->{no_whitelist};
-    if (-s $self->{conf}->{whitelist}) {  
+    if (-s $self->{conf}->{whitelist}) {
         $self->read_whitelist;
-        foreach my $sh (keys %{$self->{whitelist}}) { 
-            if ($sh ne 'sha1') { 
+        foreach my $sh (keys %{$self->{whitelist}}) {
+            if ($sh ne 'sha1') {
                 while ($headers =~ /^$sh:\s+(.*)$/img) {
                     last unless $1;
                     my $fc = $1;
                     $self->log (13,"whitelist checking headers for match $sh: $fc");
                     foreach my $address (@{$self->{whitelist}->{$sh}}) {
-                        if ($fc =~ /$address/i) { 
+                        if ($fc =~ /$address/i) {
                             $self->log (3,"ignoring mail $obj->{id}, whitelisted by rule: $sh: $address");
                             return 1;
                         }
@@ -865,17 +934,17 @@ sub local_check {
 
 
 
-sub read_whitelist { 
-    my ($self) = @_; 
+sub read_whitelist {
+    my ($self) = @_;
     return if $self->{whitelist};
 
     my %whitelist;
     my $lines = $self->read_file($self->{conf}->{whitelist},0,1);
-    for (@$lines) { 
+    for (@$lines) {
         s/^\s*//;
         next if /^#/;
         chomp;
-        my ($type, $value) = split /\s+/, $_, 2; 
+        my ($type, $value) = split /\s+/, $_, 2;
         $type =~ y/A-Z/a-z/ if $type;
         push @{$whitelist{$type}}, $value if ($type && $value);
     }
@@ -883,11 +952,11 @@ sub read_whitelist {
     $self->log (8,"loaded ". scalar(keys %whitelist) ." different types of whitelist");
     #$self->logobj (15,"loaded whitelist:", \%whitelist);
     return 1;
-}   
+}
 
 
 sub logerr {
-    my ($self,$msg) = @_; 
+    my ($self,$msg) = @_;
     $msg = $self->errstr unless $msg;
     $self->log(1,"$self->{breed} error: ". $msg);
     return;
@@ -895,27 +964,6 @@ sub logerr {
 
 
 
-sub create_symlinks {
-    my @progs = @_;
-    my $bdir = $0 =~ /^\// ? $0 : "$ENV{PWD}/$0";
-    $bdir =~ s|/([^/]+)$||;
-    foreach (@progs) {
-        if (-e "$bdir/$_") {
-            next if -l "$bdir/$_";           # unless a symlink, prolly razor2
-            rename ("$bdir/$_", "$bdir/$_.v1")  # rename razor1 
-        }
-        print "Creating symlink $1 <== $bdir/$_\n";
-        symlink($1, "$bdir/$_") if $^O ne 'VMS';
-        
-        # Do VMS specific stuff.
-        if ($^O eq 'VMS' ) {
-            my ($disk, $dir, $file) = File::Spec->splitpath($1);
-            File::Copy::copy($1,File::Spec->catpath($disk,$dir,"$_"));
-        }
-     }
-}
-
-
 # see nextserver() for explanation of how data is stored
 #
 sub get_server_info {
@@ -950,13 +998,19 @@ sub readservers {
         $self->{s}->{$lf} = [];
         foreach (@$h) {
             push @{$self->{s}->{$lf}}, $1
-                if /^(([^\.\s]+\.)+[^\.\s]+(:\S+)?)/;  
+                if /^(([^\.\s]+\.)+[^\.\s]+(:\S+)?)/;
         }
-        if (ref($self->{s}->{$lf})) {
+        if (defined($self->{s}->{$lf}) && ref($self->{s}->{$lf})) {
             $self->log(11,"Read ". scalar(@{$self->{s}->{$lf}}) ." from server listfile: ".
                 $self->{conf}->{"listfile_$lf"});
         }
     }
+    foreach my $lf (qw(discovery nomination catalogue)) {
+        next unless defined($self->{s}->{$lf});
+        next unless ref($self->{s}->{$lf});
+        next unless @{$self->{s}->{$lf}} > 1;
+        fisher_yates_shuffle($self->{s}->{$lf});
+    }
     if ($self->{breed} =~ /^check/) {
         $self->{s}->{list} = $self->{s}->{catalogue};
         $self->{s}->{listfile} = $self->{conf}->{listfile_catalogue}; # for discovery()
@@ -974,7 +1028,7 @@ sub loadservercache {
     #
     # NOTE: this reads all server.*.conf files in razor home, not just those in .lst
     #
-    
+
     # load defaults for .lst servers
     foreach (qw(nomination catalogue)) {
         next unless $self->{s}->{$_};
@@ -1022,15 +1076,15 @@ sub writeservers {
     }
     $self->log(11,"No bootstrap_discovery (DNS) recently, not recording .lst files")
         unless scalar (@{$self->{s}->{modified_lst}});
-    $self->{s}->{modified_lst} = []; 
+    $self->{s}->{modified_lst} = [];
 
     foreach (@{$self->{s}->{modified}}) {
         my $fn = "$self->{razorhome}/server.$_.conf";
         my $header = "#\n# Autogenerated by $self->{name_version}, ". localtime() ."\n";
-        $self->write_file($fn, $self->{s}->{allconfs}->{$_}, 0, $header) 
+        $self->write_file($fn, $self->{s}->{allconfs}->{$_}, 0, $header)
             || $self->debug("writeservers skipping $fn");
     }
-    $self->{s}->{modified} = []; 
+    $self->{s}->{modified} = [];
     $self->errstrrst;  # nuke error string if write errors
     return $self;
 }
